@@ -18,7 +18,7 @@ program = os.path.basename(sys.argv[0])
 
 shared_base = '/usr/share/pycentral/'
 shared_base2 = '/usr/share/pyshared/'
-pycentral_version = '0.6.15ubuntu1'
+pycentral_version = '0.6.16'
 req_pycentral_version = '0.6.15'
 
 
@@ -421,8 +421,20 @@ class DebPackage:
             if lines:
                 pass
             elif os.environ.has_key("PYCENTRAL_NO_DPKG_QUERY"):
+                import subprocess
+                # a little convoluted, but we need to get the current architecture
+                # name
+                cmd = ['/usr/bin/dpkg', '--print-architecture']
+                p = subprocess.Popen(cmd, bufsize=1,
+                                     shell=False, stdout=subprocess.PIPE)
+                fd = p.stdout
+                arch = fd.readline().strip()
+                fd.close()
                 logging.debug("Not using dpkg-query as requested")
-                lines = map(string.strip, open('/var/lib/dpkg/info/%s.list' % self.name).readlines())
+                if os.path.exists('/var/lib/dpkg/info/%s.list' % self.name):
+                    lines = map(string.strip, open('/var/lib/dpkg/info/%s.list' % self.name).readlines())
+                else:
+                    lines = map(string.strip, open('/var/lib/dpkg/info/%s:%s.list' % (self.name, arch)).readlines())
             else:
                 cmd = ['/usr/bin/dpkg-query', '-L', self.name]
                 try:
@@ -1377,6 +1389,18 @@ class Action:
     def run(self, global_opts):
         pass
 
+    def get_arch(self):
+        import subprocess
+        # a little convoluted, but we need to get the current architecture
+        # name
+        cmd = ['/usr/bin/dpkg', '--print-architecture']
+        p = subprocess.Popen(cmd, bufsize=1,
+                             shell=False, stdout=subprocess.PIPE)
+        fd = p.stdout
+        arch = fd.readline().strip()
+        fd.close()
+        return arch
+
 
 class ActionByteCompile(Action):
     """byte compile the *.py files in <package> using the the
@@ -1413,7 +1437,9 @@ class ActionByteCompile(Action):
         if not self.runtime:
             self.error("unknown runtime version %s" % self.options.version)
 
-        if not os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname):
+        arch = self.get_arch()
+        if not os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname) \
+           and not os.path.exists('/var/lib/dpkg/info/%s:%s.list' % (self.pkgname, arch)):
             self.error("package %s is not installed" % self.pkgname)
         self.pkg = DebPackage('package', self.pkgname, oldstyle=False,
                               default_runtime=self.runtime)
@@ -1469,7 +1495,9 @@ class ActionPkgInstall(Action):
             self._option_parser.print_help()
             sys.exit(1)
         self.pkgname = self.args[0]
-        if not os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname):
+        arch = self.get_arch()
+        if not os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname) \
+           and not os.path.exists('/var/lib/dpkg/info/%s:%s.list' % (self.pkgname, arch)):
             self.error("package %s is not installed" % self.pkgname)
         return self.errors_occured
 
@@ -1580,7 +1608,9 @@ class ActionPkgPrepare(Action):
                 # first matching runtime)
                 pkg.default_runtime = get_runtime_for_version(used_runtimes[0])
 
-        if os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname):
+        arch = self.get_arch()
+        if os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname) \
+           or os.path.exists('/var/lib/dpkg/info/%s:%s.list' % (self.pkgname, arch)):
             old_pkg = DebPackage('package', self.args[0], oldstyle=False)
             try:
                 old_pkg.read_version_info()
@@ -1619,7 +1649,9 @@ class ActionBCRemove(Action):
             self._option_parser.print_help()
             sys.exit(1)
         self.pkgname = self.args[0]
-        if not os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname):
+        arch = self.get_arch()
+        if not os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname) \
+           and not os.path.exists('/var/lib/dpkg/info/%s:%s.list' % (self.pkgname, arch)):
             self.error("package %s is not installed" % self.pkgname)
         return self.errors_occured
 
@@ -1645,7 +1677,9 @@ class ActionPkgRemove(Action):
             self._option_parser.print_help()
             sys.exit(1)
         self.pkgname = self.args[0]
-        if not os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname):
+        arch = self.get_arch()
+        if not os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname) \
+           and not os.path.exists('/var/lib/dpkg/info/%s:%s.list' % (self.pkgname, arch)):
             self.error("package %s is not installed" % self.pkgname)
         return self.errors_occured
 
@@ -1675,7 +1709,9 @@ class ActionPkgList(Action):
             self._option_parser.print_help()
             sys.exit(1)
         self.pkgname = self.args[0]
-        if not os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname):
+        arch = self.get_arch()
+        if not os.path.exists('/var/lib/dpkg/info/%s.list' % self.pkgname) \
+           and not os.path.exists('/var/lib/dpkg/info/%s/%s.list' % (arch, self.pkgname)):
             self.error("package %s is not installed" % self.pkgname)
         return self.errors_occured
 
@@ -1734,9 +1770,10 @@ class ActionRuntimeInstall(Action):
         for pkgname, vstring in packages:
             try:
                 requested = list(pyversions.requested_versions(vstring, version_only=True))
-                if not self.rtname_short in requested:
-                    logging.info("%s not in what requested_versions() returned, adding it anyway" % self.rtname_short)
-                    requested.append(self.rtname_short)
+#                if not self.rtname_short in requested:
+#                    logging.info("%s not in what requested_versions() returned, adding it anyway" % self.rtname_short)
+#                    requested.append(self.rtname_short)
+
             except ValueError:
                 logging.info('\tunsupported for %s: %s (%s)' % (self.rtname, pkgname, vstring))
                 continue
@@ -1793,7 +1830,8 @@ class ActionRuntimeRemove(Action):
         needed_packages = []
         import subprocess
         for pkgname, vstring in packages:
-            if not os.path.exists('/var/lib/dpkg/info/%s.list' % pkgname):
+            if not os.path.exists('/var/lib/dpkg/info/%s.list' % pkgname) \
+               and not os.path.exists('/var/lib/dpkg/info/%s:%s.list' % (pkgname, arch)):
                 # already removed, but /var/lib/dpkg/status not yet updated
                 continue
             cmd = ['/usr/bin/dpkg-query', '-W', '-f', '${Status}\n', pkgname]
@@ -1878,6 +1916,7 @@ class ActionUpdateDefault(Action):
 
 register_action(ActionUpdateDefault)
 
+
 class ActionCleanupPkgPrepareUpdates(Action):
     name = 'cleanup-pkgprepare-updates'
     help = 'cleanup dangling symlinks and byte code files'
@@ -1902,6 +1941,7 @@ class ActionCleanupPkgPrepareUpdates(Action):
         
         # then remove those not belonging to any package
         dirs = []
+        removed = 0
         for fn in unowned:
             try:
                 os.unlink(fn + 'c')
@@ -1910,7 +1950,9 @@ class ActionCleanupPkgPrepareUpdates(Action):
                 pass
             try:
                 os.unlink(fn)
-                dirs.append(os.path.dirname(fn))
+                removed += 1
+                if not os.path.dirname(fn) in dirs:
+                    dirs.append(os.path.dirname(fn))
             except OSError, e:
                 self.error("error removing dangling symlink: %s" % str(e),
                            go_on=True)
@@ -1923,7 +1965,8 @@ class ActionCleanupPkgPrepareUpdates(Action):
         if len(packaged) > 0:
             for fn in packaged:
                 self.warn('dangling symlink owned by package: %s' % fn)
-        self.warn('removed %d dangling symlinks not owned by any package')
+        if removed:
+            self.warn('removed %d dangling symlinks not owned by any package' % removed)
         return self.errors_occured
 
     def check_dangling_links(self):
@@ -1949,28 +1992,56 @@ class ActionCleanupPkgPrepareUpdates(Action):
                     dangling.append(link)
         return dangling
 
-    def links_in_packages(self, links):
+    def locate(self, arg_list, cmd_list):
         from subprocess import PIPE, Popen
-        def chunks(l, n):
-            return [l[i:i+n] for i in range(0, len(l), n)]
-        packaged = set()
-        notfound = 0
-        for arg_list in chunks(links, 50):
-            (stdout, stderr) = Popen(["dpkg","-S"] + arg_list,
-                                     env={"LANG" : "C"},
-                                     stdout=PIPE, stderr=PIPE).communicate()
-            if stdout:
-                for line in map(string.strip, stdout.split("\n")):
-                    if line == '': continue
-                    fn = line.split(':', 1)[1][1:]
-                    if fn in arg_list:
-                        packaged.add(fn)
-            if stderr:
-                for line in map(string.strip, stderr.split("\n")):
-                    if line == '': continue
-                    fn = line.split(':', 1)[1][1:-11]
-                    notfound += 1
-        return packaged
+        (stdout, stderr) = Popen(cmd_list + arg_list,
+                                 env={"LANG" : "C"},
+                                 shell=False,
+                                 stdout=PIPE, stderr=PIPE).communicate()
+        if stdout:
+            for line in map(string.strip, stdout.split("\n")):
+                if line == '': continue
+                fn = line.split(':', 1)[1][1:]
+                if fn in arg_list:
+                    self.packaged.add(fn)
+        if stderr:
+            for line in map(string.strip, stderr.split("\n")):
+                if line == '': continue
+                fn = line.split(':', 1)[1][1:-11]
+                self.notfound += 1
+
+    def links_in_packages(self, links):
+        def chunks(l):
+            if 'SC_ARG_MAX' in os.sysconf_names:
+                chunk_max = os.sysconf('SC_ARG_MAX')
+            else:
+                chunk_max = 16384
+            chunk_max -= reduce(lambda x,y: x+y,
+                                [len(k)+len(v)+3 for k,v in os.environ.items()])
+            chunk_max -= 20
+            chunk = []
+            lchunk = 0
+            all_chunks = []
+            for i in l:
+                if lchunk + len(i)+1 > chunk_max:
+                    all_chunks.append(chunk)
+                    chunk = []
+                    lchunk = 0
+                chunk.append(i)
+                lchunk += len(i) + 1
+            if chunk:
+                all_chunks.append(chunk)
+            return all_chunks
+
+        self.packaged = set()
+        self.notfound = 0
+        all_chunks = chunks(links)
+        arg_list = all_chunks[0]
+        cmd_list = ["dpkg","-S"]
+        self.locate(arg_list, cmd_list)
+        for arg_list in all_chunks[1:]:
+            self.locate(arg_list, cmd_list)
+        return self.packaged
 
 register_action(ActionCleanupPkgPrepareUpdates)
 
